package cz.jzitnik.game.generation;

import cz.jzitnik.game.entities.Block;
import cz.jzitnik.game.Game;
import cz.jzitnik.game.SpriteLoader;
import cz.jzitnik.game.entities.items.ItemBlockSupplier;
import cz.jzitnik.game.sprites.Steve;

import java.util.List;
import java.util.Random;
import java.util.concurrent.CopyOnWriteArrayList;

public class Generation {
    public static void generateWorld(Game game) {
        var world = game.getWorld();
        initializeWorld(world);

        Block steveBlock = new Block("steve", SpriteLoader.SPRITES.STEVE);
        steveBlock.setSpriteState(Steve.SteveState.FIRST);
        steveBlock.setGhost(true);
        Block steveBlock2 = new Block("steve", SpriteLoader.SPRITES.STEVE);
        steveBlock2.setSpriteState(Steve.SteveState.SECOND);
        steveBlock2.setGhost(true);
        steveBlock2.setMob(true);

        int[] terrainHeight = generateTerrain();

        game.getPlayer().setPlayerBlock1(steveBlock);
        game.getPlayer().setPlayerBlock2(steveBlock2);

        populateWorld(world, terrainHeight);
        plantTrees(world, terrainHeight);

        // Spawn player at a valid starting point
        world[terrainHeight[256] - 1][256].add(steveBlock2);
        world[terrainHeight[256] - 2][256].add(steveBlock);
    }

    private static void initializeWorld(List<Block>[][] world) {
        for (int i = 0; i < 256; i++) {
            for (int j = 0; j < 512; j++) {
                world[i][j] = new CopyOnWriteArrayList<>();
            }
        }
    }

    private static int[] generateTerrain() {
        Random random = new Random();
        int baseHeight = 120;
        int[] terrainHeight = new int[512];

        terrainHeight[0] = baseHeight;
        for (int i = 1; i < 512; i++) {
            int heightChange = random.nextInt(3) - 1;
            terrainHeight[i] = Math.max(100, Math.min(140, terrainHeight[i - 1] + heightChange));
        }

        for (int i = 2; i < 510; i++) {
            terrainHeight[i] = (terrainHeight[i - 1] + terrainHeight[i] + terrainHeight[i + 1]) / 3;
        }

        return terrainHeight;
    }

    private static void populateWorld(List<Block>[][] world, int[] terrainHeight) {
        Random random = new Random();

        for (int i = 0; i < 512; i++) {
            int hillHeight = terrainHeight[i];

            world[hillHeight][i].add(ItemBlockSupplier.getBlock("grass"));

            if (random.nextDouble() < 0.1 && !isTreeNearby(world, i, hillHeight)) {
                world[hillHeight - 1][i].add(ItemBlockSupplier.getBlock("grass_bush"));
            }

            for (int j = 1; j <= 4; j++) {
                if (hillHeight + j < 256) {
                    world[hillHeight + j][i].add(ItemBlockSupplier.getBlock("dirt"));
                }
            }

            for (int j = hillHeight + 5; j < 250; j++) {
                world[j][i].add(ItemBlockSupplier.getBlock("stone"));
            }

            world[255][i].add(new Block("bedrock", SpriteLoader.SPRITES.BEDROCK));
        }

        // Fill air blocks
        for (List<Block>[] lists : world) {
            for (List<Block> list : lists) {
                list.addFirst(new Block("air", SpriteLoader.SPRITES.AIR, true, false));
            }
        }
    }

    private static boolean isTreeNearby(List<Block>[][] world, int x, int y) {
        int radius = 3; // Check within a 3-block radius for trees
        for (int dx = -radius; dx <= radius; dx++) {
            for (int dy = -radius; dy <= radius; dy++) {
                int nx = x + dx;
                int ny = y + dy;
                if (nx >= 0 && nx < 512 && ny >= 0 && ny < 256) {
                    for (Block block : world[ny][nx]) {
                        if (block.getBlockId().equals("oak_log") || block.getBlockId().equals("oak_leaves")) {
                            return true;
                        }
                    }
                }
            }
        }
        return false;
    }


    private static void plantTrees(List<Block>[][] world, int[] terrainHeight) {
        Random random = new Random();
        for (int i = 10; i < 502; i += random.nextInt(20) + 20) {
            int treeBase = terrainHeight[i];

            if (treeBase - 3 < 0)
                continue;

            tree(world, i, treeBase);
        }
    }

    public static void tree(List<Block>[][] world, int i, int treeBase) {
        for (int j = 0; j < 3; j++) {
            if (treeBase - j >= 0) {
                world[treeBase - j - 1][i].add(ItemBlockSupplier.getBlock("oak_log"));
            }
        }

        int leafY = treeBase - 4;

        for (int layer = 0; layer < 3; layer++) {
            int size = 5 - (layer * 2);
            int offsetY = leafY - layer;

            for (int dx = -size / 2; dx <= size / 2; dx++) {
                int x = i + dx;
                int y = offsetY;

                if (x >= 0 && x < world[0].length && y >= 0) {
                    world[y][x].add(ItemBlockSupplier.getBlock("oak_leaves"));
                }
            }
        }
    }
}
